function [ U, opts ] = pdetwo( f_func, diff_h, diff_v, bndry_h, bndry_v, uinit, varargin )
% PDETWO - Solve two dimensional PDE's using ALGORITHM 565 PDETWO/PSETM/GEARB.
% 
% The types of PDE's that can be solved take the following form (in LaTeX notation)
%
% \begin{eqnarray*}
%  \frac{ \partial u_l }{ \partial t } &=& f_l( t,x,y,u_1,\ldots,u_{\mathrm{NPDE}},
%  \frac{ \partial u_1 }{ \partial x }, \ldots, \frac{ \partial u_{\mathrm{NPDE}} }{ \partial x },
%  \frac{ \partial u_1 }{ \partial y }, \ldots, \frac{ \partial u_{\mathrm{NPDE}} }{ \partial y }, \\
%  & & \frac{ \partial }{ \partial x } \left( \mathrm{DH}_{l,1} \frac{ \partial u_1 }{ \partial x } \right),\dots,\frac{ \partial }{ \partial x } \left( \mathrm{DH}_{l,\mathrm{NPDE}} \frac{ \partial u_{\mathrm{NPDE}} }{ \partial x } \right), \\ %\nonumber
%  & & \frac{ \partial }{ \partial y } \left( \mathrm{DV}_{l,1} \frac{ \partial u_1 }{ \partial y } \right),\dots,\frac{ \partial }{ \partial y } \left( \mathrm{DV}_{l,\mathrm{NPDE}} \frac{ \partial u_{\mathrm{NPDE}} }{ \partial y } \right) ) %\quad \mathrm{for} \quad l=1,2,\ldots,\mathrm{NPDE}
%  %\nonumber
% \end{eqnarray*}
% 
% With u a vector of NPDE unknown functions, u_x and u_y representing the 
% x and y spatical derivatives respectfully.  The "Y" boundary conditions 
% take the form of equations with the following form
% 
% \mathrm{AH}_l u_l + \mathrm{BH}_l \frac{ \partial u_l }{ \partial y } = \mathrm{CH}_l
% 
% while the "X" boundary conditions take the form of equations with the 
% following form
% 
% \mathrm{AV}_l u_l + \mathrm{BV}_l \frac{ \partial u_l }{ \partial x } = \mathrm{CV}_l
% 
% The boundary conditions need NOT be consistent with the initial conditions.  The 
% initial conditions are given by NPDE "functions" in the obvious form
% 
% u_l(t_0,x,y) = \phi_l(x,y) \quad \mathrm{for} \quad (x,y) \in R, \quad l=1,2,\ldots,\mathrm{NPDE}
% 
% The user of this routine must at specify the following octave functions:
% 
% A function to evaluate the right hand side f function (defined above)
% " """""""" "" """""""" """ horizontal diffusion coefficients DH(x,t,u)
% " """""""" "" """""""" """ vertical   diffusion coefficients DV(x,t,u)
% " """""""" "" """""""" """ horizontal boundary condition function AH(x,t,u), BH(x,t,u), and CH(x,t,u)
% " """""""" "" """""""" """ vertical   """""""" """"""""" """""""" AV(x,t,u), BV(x,t,u), and CV(x,t,u)
% A function to evaluate and return the initial conditions
% 
% The integration method is based on the method of lines (MOL).  The spatial 
% discretization is based on a finite difference discretization of the 
% two dimensional domain while the temporal integration is performed with the 
% GEARB ODE package.  The GEARB ODE package is nice in that it has methods for 
% integration both stiff and non-stiff problems.  Many more details can be
% found in the original article:
% 
% Algorithm 565: PDETWO/PSETM/GEARB: Solution of Systems of Two-Dimensional Nonlinear Partial Differential
% by David K. Melgaard, and Ricard F. Sincovec.
% ACM Transactions on Mathematical Software (TOMS), Volume 7 Issue 1 (March 1981)
% Pages: 126 - 135
% 
% Written by:
% -- 
% John L. Weatherwax                2006-03-05
% 
% email: wax@alum.mit.edu
% 
% Please send comments and especially bug reports to the
% above email address.
% 
%-----

if( nargin < 6 ) error( 'Not enough input arguments' ); end

% Default pdetwo input arguments: 
%  

% The initial time:
%
t0 = 0.0;

% The left end of the computational domain:
% 
a1 = 0;        

% The right end of the computational domain:
% 
b1 = 1;        

% The bottom end of the computational domain:
%
a2 = 0; 

% The top end of the computational domain:
% 
b2 = 1; 

% The number of intervals to divide the domain into 
% we use equal sized intervals in both x & y:
% 
nint = 30; 

% The initial stepsize in time:
% 
h = 1e-7;     

% The desired relative accuracy of the ODE solver:
% 
eps = 1e-4;     

% The array of breakpoints (not required to be uniformly spaced):
xbkpt = linspace( a1, b1, nint+1 ); 
ybkpt = linspace( a2, b2, nint+1 ); 

% Output times the solution is desired:
% 
tGrid = [ t0 t0+1.0 ]; 

% Algorithm Method Flags:
% 
% MF = 10*METH + MITER

% METH  = 1 Adams method (a generalization of Crank-Nicolson)
%       = 2 backwards differentiation formulas (bdf) or stiff method of Gear
% 
METH = 2; 

% For the solution of the backwards difference equations (for the implicit methods)
% MITER = 0 means functional iteration
%       = 1 chord method with analytic jacobians, user supplies subroutines for the derivative evaluation
%           this method is in pdetwo but should be avoided
%       = 2 chord method for calculating the jacobian (calculated via the special function PSETM)
%       = 3 chord method for calculating the jacobian (with a diagonal approximation)
%           this method is in pdetwo but is generally not useful in solving PDE's
%
MITER = 2; 

mf = 10*METH + MITER;

% MORDER, THE MAXIMUM ORDER OF THE METHOD USED IN THE MODIFIED GEARB.  
% If METH = 1, IT MUST BE LESS THAN OR EQUAL TO 12.
% IF METH = 2, IT MUST BE LESS THAN OR EQUAL TO 5.  
%
morder = 5;

% Override default arguments with "varargin" inputs (if any):
%
i=1; 
while i < length(varargin)
  arg = varargin{i};
  switch lower(arg)
   case 't0'
    t0 = varargin{i+1}; 
   case { 'xl', 'a1' }
    a1 = varargin{i+1};
   case { 'xr', 'b1' }
    b1 = varargin{i+1};
   case { 'yl', 'a2' }
    a2 = varargin{i+1};
   case { 'yr', 'b2' }
    b2 = varargin{i+1};
   case 'nint'
    nint = varargin{i+1}; 
    xbkpt = linspace( a1, b1, nint+1 ); 
    ybkpt = linspace( a2, b2, nint+1 ); 
   case { 'epsilon', 'eps' } 
    eps = varargin{i+1};
   case { 'h', 'dt' }
    h = varargin{i+1};
   case { 'xbkpt', 'xgrid' }
    xbkpt = varargin{i+1};
   case { 'ybkpt', 'ygrid' }
    ybkpt = varargin{i+1};
   case 'tgrid'
    tGrid = varargin{i+1};
   case 'mf'
    mf = varargin{i+1};
   case 'ode_solver'
    switch lower(varargin{i+1})
     case 'adams'
      METH=1; 
     case 'bdf'
      METH=2; 
     otherwise
      error( 'Unknown argument to ode_solver' ); 
    end
    mf = 10*METH + MITER;
   case 'morder'
    morder = varargin{i+1};
   otherwise
    error( [ 'Unknown input: ', arg ] ); 
  end
  i=i+2;
end

% Input argument error checking:
%
% *X* domain: 
if( a1 >= b1 ) 
  error( 'a1 >= b1' ); 
end
if( (a1 ~= xbkpt(1)) | (b1 ~= xbkpt(end)) )
  error( 'a1 ~= xbkpt(1) OR b1 ~= xbkpt(end)' ); 
end
% *Y* domain: 
if( a2 >= b2 ) 
  error( 'a2 >= b2' ); 
end
if( (a2 ~= ybkpt(1)) | (b2 ~= ybkpt(end)) )
  error( 'a2 ~= ybkpt(1) OR b2 ~= ybkpt(end)' ); 
end
if( eps <= 0 )
  error( 'eps <= 0' ); 
end
if( nint < 1 )
  error( 'nint < 1' ); 
end
if( length(tGrid) < 2 ) 
  error( 'length(tGrid) < 2' ); 
end
if( length(xbkpt) > length(ybkpt) ) 
  warning( 'Better memory usage results if the problem can be defined such that NX <= NY' ); 
end

% Create a matrix of initial condtions (U0): 
%
if( is_matrix( uinit ) )      % initial conditions are entered as a matrix ...
  U0 = uinit; 
else                          % initial conditions are entered as a function ...
  U0 = feval( uinit, xbkpt, ybkpt ); 
end

% Call the octave *.oct file that calls the FORTRAN routine pdetwo:
%
U = pdetwog(tGrid,h,xbkpt,ybkpt,eps,mf,morder,...
	    f_func, diff_h, diff_v, bndry_h, bndry_v, U0 );

% Collect all inputs together into a structure:
%
opts.t0             = t0;
opts.a1             = a1;
opts.b1             = b1; 
opts.a2             = a2;
opts.b2             = b2;
opts.nint           = nint; 
opts.h              = h;
opts.eps            = eps;
opts.xbkpt          = xbkpt;
opts.ybkpt          = ybkpt;
opts.tGrid          = tGrid;
opts.mf             = mf;
opts.mf             = morder;
